from inoon_lora_packet.packet import Packet, HexConverter


class ErrorV2Packet(Packet):
    code_map = {
        1: 'OverReTx',
        2: 'JoinFail',
        3: 'Rx',
        4: 'Nok',
        5: 'OverReTxInifinite',
    }

    def _field_spec(self):
        return [
            {'name': 'code',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'rssi',
             'bytes': 1,
             'convert': HexConverter.hex_to_int8,
             'restrict': None},
        ]

    def __str__(self):
        msg = ''
        msg += 'ERR | '
        msg += 'Code: {} | '.format(self.code_map[self.code])
        msg += 'RSSI: {}'.format(self.rssi)
        return msg

    @classmethod
    def encode(cls, code, rssi):
        enc = ''
        enc += format(code, '02x')
        enc += HexConverter.int_to_hex(rssi, 1, True)
        return enc.lower()



from inoon_lora_packet.packet import (Packet, InvalidPacketError, HexConverter)


class AccWaveV3Packet(Packet):
    axis_ctrl = {
        0: ['x', 'y', 'z'],
        1: ['x'],
        2: ['y'],
        3: ['z']
    }

    range_desc = {
        0: '2G',
        1: '4G',
        2: '8G',
        3: '16G',
    }

    def _field_spec(self):
        return [
            {'name': 'ctrl',
             'bytes': 1,
             'bit_fields': [
                 {'name': 'range', 'bits': 2, 'restrict': [0, 1, 2, 3]},
                 {'name': 'axis', 'bits': 2, 'restrict': [0, 1, 2, 3]},
                 {'name': 'id', 'bits': 4, 'restrict': None}
             ]},
            {'name': 'ctrl2',
             'bytes': 1,
             'bit_fields': [
                 {'name': 'pack_type', 'bits': 4, 'restrict': [0x0, 0x1, 0xF]},
                 {'name': 'seq', 'bits': 4, 'restrict': range(0, 16)},
             ]},
        ]

    def __init__(self, raw_packet):
        super(self.__class__, self).__init__(raw_packet)

        log_packet = raw_packet[4:]
        axises = self.axis_ctrl[self.ctrl.axis]

        bin_str = ''
        for c in log_packet:
            bin_str += '{0:04b}'.format(int(c, base=16))

        bit_size = 16
        scale = 1
        if self.ctrl2.pack_type == 0x1:
            bit_size = 12
            scale = 4

        if len(bin_str) % bit_size != 0:
            raise InvalidPacketError

        frame_max = (1 << bit_size-1) - 1
        frame_count = int(len(bin_str) / bit_size)
        frame_count = int(frame_count / len(axises))

        for axis in axises:
            setattr(self, axis, [])

        for i in range(0, len(bin_str), bit_size * len(axises)):
            offset = 0
            if 'x' in axises:
                frame_str = bin_str[i+offset:i+offset+bit_size]
                parsed_value = int(frame_str, base=2)
                if parsed_value > frame_max:
                    parsed_value = parsed_value - (1 << bit_size)

                getattr(self, 'x').append(parsed_value*scale)
                offset += bit_size

            if 'y' in axises:
                frame_str = bin_str[i+offset:i+offset+bit_size]
                parsed_value = int(frame_str, base=2)
                if parsed_value > frame_max:
                    parsed_value = parsed_value - (1 << bit_size)

                getattr(self, 'y').append(parsed_value*scale)
                offset += bit_size

            if 'z' in axises:
                frame_str = bin_str[i+offset:i+offset+bit_size]
                parsed_value = int(frame_str, base=2)
                if parsed_value > frame_max:
                    parsed_value = parsed_value - (1 << bit_size)

                getattr(self, 'z').append(parsed_value*scale)
                offset += bit_size

    def __str__(self):
        msg = ''
        msg += 'WAVE | '
        msg += 'Rng: {} | '.format(self.range_desc[self.ctrl.range])
        msg += 'Axis: {} | '.format(' '.join(self.axis_ctrl[self.ctrl.axis]))
        msg += 'ID: {} | '.format(self.ctrl.id)

        if self.ctrl2.pack_type == 0x1:
            msg += 'Pack: {}(b) | '.format(12)
        elif self.ctrl2.pack_type == 0x0 or self.ctrl2.pack_type == 0xF:
            msg += 'Pack: {}(b) | '.format(16)
        else:
            msg += 'Pack: -- | '

        if self.ctrl2.seq == 0xF:
            msg += 'Seq: Fin.'
        else:
            msg += 'Seq: {}'.format(self.ctrl2.seq)

        return msg

    @classmethod
    def encode(cls, pckt_id, seq, axis, rng, values):
        """values parameter is dict type.
        """
        enc = ''

        ctrl = ((rng << 2 | axis) << 4) | pckt_id
        enc += '{:02X}'.format(ctrl)
        enc += '{:02X}'.format(seq)

        if axis == 0:
            acc_values = list(zip(values['x'], values['y'], values['z']))

            for value in acc_values:
                enc += HexConverter.int_to_hex(value[0], 2, True)
                enc += HexConverter.int_to_hex(value[1], 2, True)
                enc += HexConverter.int_to_hex(value[2], 2, True)
        else:
            axis_name = ''
            if axis == 1:
                axis_name = 'x'
            elif axis == 2:
                axis_name = 'y'
            else:
                axis_name = 'z'

            for value in values[axis_name]:
                enc += HexConverter.int_to_hex(value, 2, True)

        return enc.lower()



from inoon_lora_packet.packet import (Packet, LogPosition,
                                      HexConverter, InvalidPacketError)


class DataLogV2Packet(Packet):
    spec = [
        {'name': 'idx',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'pos',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [LogPosition[k] for k in LogPosition._member_map_]},
    ]

    def _field_spec(self):
        return self.spec

    def __init__(self, raw_packet):
        super(self.__class__, self).__init__(raw_packet)

        # TODO: Refactoring parse method.
        data_cnt = len(raw_packet[4:]) - 2

        if data_cnt % 2 is not 0:
            raise InvalidPacketError

        if data_cnt % 12 is not 0:
            raise InvalidPacketError

        self.x = []
        self.y = []
        self.z = []

        for start_idx in range(0, data_cnt, 12):
            parse_log = raw_packet[4:][start_idx:start_idx+12]
            self.x.append(HexConverter.hex_to_int16(parse_log[0:4]))
            self.y.append(HexConverter.hex_to_int16(parse_log[4:8]))
            self.z.append(HexConverter.hex_to_int16(parse_log[8:]))

        self.rssi = HexConverter.hex_to_int8(raw_packet[-2:])

    def __str__(self):
        msg = ''
        msg += 'LOG | '
        msg += 'ID: {} | '.format(self.idx)

        if self.pos == 0:
            msg += 'Alarm'
        elif self.pos == 1:
            msg += 'Mid.'
        elif self.pos == 2:
            msg += 'Finish'
        else:
            msg += '--'
        msg += ' | '

        msg += 'RSSI: {}'.format(self.rssi)
        return msg

    @classmethod
    def encode(self, idx, pos, data_cnt, x, y, z, rssi):
        enc = ''
        enc += format(idx, '02x')
        enc += format(pos, '02x')
        for i in range(0, data_cnt):
            enc += HexConverter.int_to_hex(x[i], 2, True)
            enc += HexConverter.int_to_hex(y[i], 2, True)
            enc += HexConverter.int_to_hex(z[i], 2, True)

        enc += HexConverter.int_to_hex(rssi, 1, True)

        return enc.lower()



from inoon_lora_packet.packet import Packet, HexConverter, ErrorType


class ErrorV3Packet(Packet):
    def _field_spec(self):
        return [
            {'name': 'err_type',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [ErrorType[k] for k in ErrorType._member_map_]},
        ]

    def __str__(self):
        code_desc = {
            ErrorType.lora_over_re_tx: 'OverReTx',
            ErrorType.lora_join_fail: 'JoinFail',
            ErrorType.lora_rx: 'RxFail',
            ErrorType.lora_nok: 'Nok',
            ErrorType.lora_tx_inf: 'TxInfinite',
            ErrorType.lora_skip_high_g: 'SkipHighG',
            ErrorType.lora_flash_store: 'Flash Store failed',
            ErrorType.lora_flash_load: 'Flash Load failed',
            ErrorType.lora_flash_clear: 'Flash Clear failed',
            ErrorType.lora_flash_update: 'Flash Update failed',
            ErrorType.lora_power: 'LoRa Power failed',
            ErrorType.lora_beacon_start: 'ADV. TX failed',
            ErrorType.lora_fw_check: 'SoluM FW fail',
            ErrorType.mems_check: 'BMA280 err',
        }

        msg = ''
        msg += 'ERR | '
        msg += 'Type: {} | '.format(code_desc[self.err_type])
        return msg

    @property
    def code(self):
        """For V2 compatibility"""
        return self.err_type

    @classmethod
    def encode(cls, err_type):
        return format(err_type, '02x').lower()



from inoon_lora_packet.packet import Packet, HexConverter


class ReportV3Packet(Packet):
    def _field_spec(self):
        return [
            {'name': 'ctrl',
             'bytes': 1,
             'bit_fields': [
                 {'name': 'method', 'bits': 1, 'restrict': None},
                 {'name': 'overflow', 'bits': 1, 'restrict': None},
                 {'name': 'resv', 'bits': 2, 'restrict': None},
                 {'name': 'scale', 'bits': 4, 'restrict': None}
             ]},
            {'name': 'avg',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},
            {'name': 'stddev',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},
            {'name': 'min',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},
            {'name': 'max',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},
        ]

    def __str__(self):
        msg = ''
        msg += 'REPORT | '
        msg += 'Scale: {} | '.format(self.ctrl.scale)
        msg += 'Avg.: {}mg | '.format(self.gravity(self.avg))
        msg += 'StdDev.: {}mg | '.format(self.gravity(self.stddev))
        msg += 'Min.: {}mg | '.format(self.gravity(self.min))
        msg += 'Max.: {}mg '.format(self.gravity(self.max))
        return msg

    def gravity(self, value):
        return round(float(value) / (2 ** self.ctrl.scale) * 3.91, 2)

    @classmethod
    def encode(cls, scale, avg, std, min_val, max_val):
        enc = ''
        enc += '{:02X}'.format(scale)
        enc += HexConverter.int_to_hex(avg, 2, True)
        enc += HexConverter.int_to_hex(std, 2, False)
        enc += HexConverter.int_to_hex(min_val, 2, True)
        enc += HexConverter.int_to_hex(max_val, 2, True)
        return enc.lower()



from inoon_lora_packet.packet import (Packet, NoticeType,
                                      HexConverter, InvalidPacketError)


class NoticeV2Packet(Packet):
    def _field_spec(self):
        try:
            notice_type = int(self.raw_packet[:2], 16)

            if notice_type == NoticeType.power_up:
                return PowerUpV2Packet.spec
            elif notice_type == NoticeType.power_off:
                return PowerOffV2Packet.spec
            elif notice_type == NoticeType.setup:
                return SetupV2Packet.spec

            raise InvalidPacketError
        except Exception:
            raise InvalidPacketError

    def __str__(self):
        msg = None
        if self.type == NoticeType.power_up:
            msg = self.__power_up_msg()
        elif self.type == NoticeType.power_off:
            msg = self.__power_off_msg()

        elif self.type == NoticeType.setup:
            msg = self.__setup_msg()

        msg += 'RSSI: {}'.format(self.rssi)
        return msg

    def __power_up_msg(self):
        msg = 'ON | '
        msg += 'Reset({}): '.format(self.reason)
        if self.reason & 0x08 == 0x08:
            msg += 'CPU '
        if self.reason & 0x04 == 0x04:
            msg += 'SW '
        if self.reason & 0x02 == 0x02:
            msg += 'WD '
        if self.reason & 0x01 == 0x01:
            msg += 'nRST'
        if self.reason is 0:
            msg += '--'
        msg += ' | '

        msg += 'Err({}): '.format(self.error)
        if self.error & 0x04 == 0x04:
            msg += 'Join '
        if self.error & 0x02 == 0x02:
            msg += 'TX '
        if self.error & 0x01 == 0x01:
            msg += 'Alive'
        if self.error is 0:
            msg += '--'
        msg += ' | '

        return msg

    def __power_off_msg(self):
        msg = 'OFF | '
        if self.data == 1:
            msg += 'LoRa Off Cmd'
        elif self.data == 2:
            msg += 'BLE Off Cmd'
        elif self.data == 3:
            msg += 'Uninstalled Dev(Auto Off)'
        elif self.data == 4:
            msg += 'Low BAT'
        elif self.data == 5:
            msg += 'Error'
        else:
            msg += ''
        msg += ' | '

        return msg

    def __setup_msg(self):
        msg = 'SETUP | '
        if self.install == 0:
            msg += 'Uninstall'
        else:
            msg += 'Install'
        msg += ' | '
        return msg


class PowerUpV2Packet():
    spec = [
        {'name': 'type',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'len',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [4]},

        {'name': 'reason',
         'bytes': '1',
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'resv',
         'bytes': '2',
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'error',
         'bytes': '1',
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'rssi',
         'bytes': 1,
         'convert': HexConverter.hex_to_int8,
         'restrict': None},
    ]

    @classmethod
    def encode(cls, reason, error, rssi):
        enc_val = ''
        enc_val += format(NoticeType.power_up, '02x')
        enc_val += format(4, '02x')
        enc_val += format(reason, '02x')
        enc_val += '0000'
        enc_val += format(error, '02x')
        enc_val += HexConverter.int_to_hex(rssi, 1, True)
        return enc_val.lower()


class PowerOffV2Packet():
    spec = [
        {'name': 'type',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'len',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [1]},

        {'name': 'data',
         'bytes': '1',
         'convert': HexConverter.hex_to_uint,
         'restrict': [1, 2, 3, 4, 5]},

        {'name': 'rssi',
         'bytes': 1,
         'convert': HexConverter.hex_to_int8,
         'restrict': None},
    ]

    @classmethod
    def encode(cls, data, rssi):
        enc_val = ''
        enc_val += format(NoticeType.power_off, '02x')
        enc_val += format(1, '02x')
        enc_val += format(data, '02x')
        enc_val += HexConverter.int_to_hex(rssi, 1, True)
        return enc_val.lower()


'''
class ResetV2Packet(Packet):
    def _field_spec(self):
        return [
            {'name': 'type',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'len',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [0]},
        ]
'''


class SetupV2Packet(Packet):
    spec = [
        {'name': 'type',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'len',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [1]},

        {'name': 'install',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [0, 1]},

        {'name': 'rssi',
         'bytes': 1,
         'convert': HexConverter.hex_to_int8,
         'restrict': None},
    ]

    @classmethod
    def encode(cls, install, rssi):
        enc_val = ''
        enc_val += format(NoticeType.setup, '02x')
        enc_val += format(1, '02x')
        enc_val += format(install, '02x')
        enc_val += HexConverter.int_to_hex(rssi, 1, True)
        return enc_val.lower()



class BMA250E():
    mg_per_step = {
        1: 3.91,
        2: 7.81,
        3: 15.63,
        4: 31.25,
    }

    @classmethod
    def gravity(cls, sensitivity, raw_value):
        return round(cls.mg_per_step[sensitivity] * raw_value / 1000, 5)


class BMA280():
    mg_per_step = {
        1: 0.244,
        2: 0.488,
        3: 0.977,
        4: 1.953,
    }

    @classmethod
    def gravity(cls, sensitivity, raw_value):
        return round(cls.mg_per_step[sensitivity] * raw_value / 1000, 5)



from setuptools import setup


setup_requires = [
]

install_requires = [
]

dependency_links = [
]

setup(
    name='inoon_lora_packet',
    version='5.0.0',
    description='LoRa packet parser for Ino-on, Inc.',
    author='Joonkyo Kim',
    author_email='jkkim@ino-on.com',
    packages=['inoon_lora_packet', 'inoon_lora_packet.base',
              'inoon_lora_packet.bsp', 'inoon_lora_packet.v2',
              'inoon_lora_packet.v3'],
    include_package_data=True,
    install_requires=install_requires,
    setup_requires=setup_requires,
    dependency_links=dependency_links,
    # scripts=['manage.py'],
    entry_points={
        'console_scripts': [
        ],
        "egg_info.writers": [
            "foo_bar.txt = setuptools.command.egg_info:write_arg",
        ],
    },
)



from inoon_lora_packet.packet import (PacketType, PacketHeaderV3,
                                      InvalidPacketError,
                                      NotSupportedPacketError)
from inoon_lora_packet.base import base_parser
from .ack import AckV3Packet
from .alive import AliveV3Packet
from .event import EventV3Packet
from .error import ErrorV3Packet
from .notice import NoticeV3Packet
from .data_log import DataLogV3Packet
from .report import ReportV3Packet
from .acc_wave import AccWaveV3Packet
from .inclination import InclinationV3Packet
from .mr_measurement import MRMeasurementV3Packet
from .mr_report import MRReportV3Packet


class Ver3Parser(base_parser.Parser):
    @property
    def header_length(self):
        return 20

    def parse(self, raw_packet):
        if len(raw_packet) < self.header_length:
            raise InvalidPacketError

        header = PacketHeaderV3(raw_packet)
        raw_payload = raw_packet[self.header_length:-2]
        packet = None

        if header.packet.type == PacketType.alive:
            packet = AliveV3Packet(raw_payload)
        elif header.packet.type == PacketType.notice:
            packet = NoticeV3Packet(raw_payload)
        elif header.packet.type == PacketType.data_log:
            packet = DataLogV3Packet(raw_payload)
        elif header.packet.type == PacketType.event:
            packet = EventV3Packet(raw_payload)
        elif header.packet.type == PacketType.ack:
            packet = AckV3Packet(raw_payload)
        elif header.packet.type == PacketType.error:
            packet = ErrorV3Packet(raw_payload)
        elif header.packet.type == PacketType.report:
            packet = ReportV3Packet(raw_payload)
        elif header.packet.type == PacketType.acc_wave:
            packet = AccWaveV3Packet(raw_payload)
        elif header.packet.type == PacketType.inclination:
            packet = InclinationV3Packet(raw_payload)
        elif header.packet.type == PacketType.mr_measurement:
            packet = MRMeasurementV3Packet(raw_payload)
        elif header.packet.type == PacketType.mr_report:
            packet = MRReportV3Packet(raw_payload)
        else:
            raise NotSupportedPacketError

        return header, packet
